package raincoat.builder;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import backtype.storm.Config;
import backtype.storm.generated.StormTopology;
import backtype.storm.spout.SchemeAsMultiScheme;
import backtype.storm.topology.BoltDeclarer;
import backtype.storm.topology.IRichBolt;
import backtype.storm.topology.IRichSpout;
import backtype.storm.topology.TopologyBuilder;
import backtype.storm.tuple.Fields;
import backtype.storm.tuple.Tuple;
import raincoat.api.Operation;
import raincoat.api.RATreeNode;
import raincoat.api.RaincoatGrouping;
import raincoat.kafka.RaincoatScheme;
import raincoat.operations.Projection;
import raincoat.operations.Selection;
import raincoat.operations.Sum;
import raincoat.ratree.AggregatorNode;
import raincoat.ratree.AggregatorNode.AggregatorType;
import raincoat.ratree.JoinNode;
import raincoat.ratree.ProjectNode;
import raincoat.ratree.SelectNode;
import raincoat.ratree.SourceNode;
import raincoat.ratree.WindowNode;
import raincoat.sql.AlgebraOperator;
import raincoat.sql.Condition;
import raincoat.storm.FunctionBolt;
import raincoat.storm.HashJoinBolt;
import raincoat.storm.Tuples;
import raincoat.storm.WindowedBolt;
import raincoat.test.generators.RandomIntegerSpout;
import raincoat.testsuit.performance.DataGenSpout;
import raincoat.testsuit.performance.DataGenerator;
import raincoat.testsuit.performance.FacebookDataGenerator;
import raincoat.testsuit.performance.InstagramDataGenerator;
import raincoat.testsuit.performance.LoggingKafkaSpout;
import raincoat.testsuit.performance.TwitterDataGenerator;
import raincoat.util.DFS;
import raincoat.windows.SlidingTimedWindow;
import storm.kafka.SpoutConfig;
import storm.kafka.KafkaConfig.StaticHosts;


public class RaincoatFactory extends DFS {

	private TopologyBuilder _builder;
	private WindowNode _windowNode;

	public RaincoatFactory() {
	}

	public StormTopology makeTopology(RATreeNode root) {
		if (!(root instanceof WindowNode)) {
			throw new IllegalArgumentException(
					"The root of the RA tree must be a WindowNode");
		}
		else if (root.getChildren().size() != 1) {
			throw new IllegalArgumentException(
					"The WindowBolt cannot have more or less than " +
					"one child.");
		}

		_builder = new TopologyBuilder();

		_windowNode = (WindowNode) root;
		traverserTree(root.getChildren().get(0));
		return _builder.createTopology();
	}

	@Override
	protected void execute(RATreeNode root) {		
		if (root instanceof SourceNode) {
			SourceNode sourceNode = (SourceNode) root;
			makeSpout(sourceNode);
		} 
		else if (root instanceof SelectNode
				|| root instanceof ProjectNode) {
			makeFunctionBolt(root);
		} 
		else if (root instanceof JoinNode) {
			JoinNode joinNode = (JoinNode) root;
			makeJoinBolt(joinNode);

		}
		else if (root instanceof AggregatorNode) {
			AggregatorNode aggrNode = (AggregatorNode) root;
			makeAggregatorBolt(aggrNode);
		}


		if (root.needsPostProcessing()) {
			RATreeNode postProcessingNode = root.getPostProcessingNode();
			postProcessingNode.addChild(root);
			execute(postProcessingNode);
		}

	}	

	private void makeSpout(SourceNode node) {
		IRichSpout spout = spoutForType(node);
		_builder.setSpout(node.getIdentifier(), spout, 
				node.getParallelismHint());
//				8);
	}

	private void makeFunctionBolt(RATreeNode node) {
		Operation<Tuple> operation = makeOperation(node);
		IRichBolt bolt = new FunctionBolt(operation);
		subscribeToChildren(node, 
				_builder.setBolt(node.getIdentifier(), 
						bolt, node.getParallelismHint()));
	}

	private void makeJoinBolt(JoinNode joinNode) {
		IRichBolt hashJoin = new HashJoinBolt(
				joinNode.getCondition(), 
				new Fields(joinNode.getOutputTuples()), 
				joinNode.getChildren().get(0).getIdentifier(), 
				joinNode.getChildren().get(1).getIdentifier(),
				windowLifespan(),  
				getTickSize(), 
				new Fields(joinNode.getOutputTuples()));

		subscribeToChildren(joinNode, 
				_builder.setBolt(joinNode.getIdentifier(), 
						hashJoin, joinNode.getParallelismHint()));
	}

	private void makeAggregatorBolt(AggregatorNode aggrNode) {
		IRichBolt bolt = new WindowedBolt(
				makeAggregatorOperator(aggrNode), 
				new SlidingTimedWindow(
						windowLifespan(),
						getTickSize()));

		subscribeToChildren(aggrNode, 
				_builder.setBolt(aggrNode.getIdentifier(), 
						bolt, aggrNode.getParallelismHint()));
	}


	private void subscribeToChildren(RATreeNode root, BoltDeclarer boltDeclarer) {
		for (RATreeNode child : root.getChildren()) {
			if(root.getGrouping() == RaincoatGrouping.SHUFFLE) {
				boltDeclarer.shuffleGrouping(child.getIdentifier());
				continue;
			} 
			else if (root.getGrouping() == RaincoatGrouping.FIELDS) {
				// TODO: Nodes using FIELDS grouping should inherit from an FieldsGroupingNode. Such that the fields grouping don't have to be specific. e.g. like it is for join nodes
				// assuming that the node is a JoinNode for now
				JoinNode joinNode = (JoinNode) root;
				Fields fields = new Fields();
				Condition condition = joinNode.getCondition();
				if (child.getOutputTuples()
						.contains(
								(String)condition
								.getFieldA()
								.getValue())) {
					fields = new Fields(
							(String)condition
							.getFieldA()
							.getValue());
				}
				else {
					fields = new Fields(
							(String)condition
							.getFieldB()
							.getValue());
				}
				boltDeclarer.fieldsGrouping(child.getIdentifier(), fields);
				continue;
			}
			throw new IllegalArgumentException(
					"Grouping ("+root.getGrouping()+") not mapped.");
		}
	}


	/* xxx hack! 
	 * reflection is the configurable way to go
	 */
	private IRichSpout spoutForType(SourceNode node) {
		String type = node.getSourceType();
		Map<String, String> options = node.getOptions();
		if (type.equals("kafka") 
				|| type.toLowerCase().equals("redisstream")) {
			
			// TODO. fix to use options
			List<String> hostStrings = new ArrayList<String>();
			hostStrings.add(options.get("host"));

			SpoutConfig config = new SpoutConfig(
					StaticHosts.fromHostString(
							hostStrings, Integer.parseInt(options.get("partitions"))), 
							options.get("topic"), 
							"/"+options.get("zkroot"), 
							options.get("id"));
			config.scheme = new SchemeAsMultiScheme(new RaincoatScheme( 
					new Fields(node.getOutputTuples())));
			
			if (options.containsKey("offset")) {
				int offset = Integer.parseInt(options.get("offset"));
				config.forceStartOffsetTime(offset);
			}
			
			
			IRichSpout kafkaSpout = new LoggingKafkaSpout(config, 10000);
			return kafkaSpout;
		} 
		else if (type.equals("random-int") || type.equals("randomint")) {
			int from = Integer.parseInt(options.get("from"));
			int to = Integer.parseInt(options.get("to"));
			int rate = 1;
			if (options.containsKey("rate")) {
				rate = Integer.parseInt(options.get("rate"));
			}
			
			return new RandomIntegerSpout(from, to, rate,
					new Fields(node.getOutputTuples()));
			
		} else if (type.equalsIgnoreCase("gen")) {
			long everyMS = Long.parseLong(options.get("every"));
			long seed = Long.parseLong(options.get("seed"));
			String genType = options.get("type");
			DataGenerator datagen = getGeneratorForType(genType, seed);
			
			return new DataGenSpout(datagen, 
					new Fields(node.getOutputTuples()), 
					everyMS);
		}

		throw new IllegalArgumentException(
				"Spout type not mapped: " + type);

	}
	
	public DataGenerator getGeneratorForType(String genType, long seed) {
		DataGenerator datagen;
		if (genType.equalsIgnoreCase("fb")) {
			datagen = new FacebookDataGenerator(seed); 
		} else if (genType.equalsIgnoreCase("twitter")) {
			datagen = new TwitterDataGenerator(seed);
		} else {
			datagen = new InstagramDataGenerator(seed);
		}
		return datagen;
	}

	private Operation<Tuple> makeOperation(
			RATreeNode node) {
		if (node.getOperator() == AlgebraOperator.SELECT) {
			return new Selection(
					new Fields(node.getOutputTuples()),
					((SelectNode) node ).getConditions());
		}
		if (node.getOperator() == AlgebraOperator.PROJECT) {
			return new Projection(new Fields(node.getOutputTuples()));
		}
		throw new IllegalArgumentException(
				"Operator not mapped.");
	}

	private Operation<Tuples> makeAggregatorOperator(AggregatorNode node) {
		if (node.getType() == AggregatorType.SUM) {
			return new Sum((String)node.getField().getValue());
		}
		throw new IllegalArgumentException(
				"Aggregator not mapped.");
	}

	private Integer windowLifespan() {
		return _windowNode.getWindowSize();
	}

	private Integer getTickSize() {
		return _windowNode.getTickInterval();
	}

	public static void main(String[] ignore) {
//		Config conf = new Config();
//		conf.setDebug(true);
//		conf.setNumWorkers(2);
//		conf.put("topology.tick.tuple.freq.secs", 2);
//
//		List<String> hostStrings = new ArrayList<String>();
//		hostStrings.add("localhost");
//		conf.put("kafka.hosts", hostStrings);
//		conf.put("kafka.partitions", 1);
//		conf.put("kafka.zkroot", "/kafka-spout");
//
//		RaincoatFactory topologyFactory = new RaincoatFactory(conf);
//
//		Set<String> sourceOutputFields = new HashSet<String>();
//		sourceOutputFields.add("seq");
//		sourceOutputFields.add("integer");
//		RATreeNode sourceNode = new SourceNode(
//				null, 
//				sourceOutputFields, 
//				"tester", "kafka", null);
//
//		List<Condition> conditions = new ArrayList<Condition>();
//		conditions.add(new Condition(new FieldValue(5, true), 
//				new FieldValue("integer", false), 
//				Operator.GREATER_THAN));
//		RATreeNode selectNode = new SelectNode(sourceOutputFields, sourceOutputFields, conditions);
//		selectNode.addChild(sourceNode);
//
//		Set<String> sourceOutputFields2 = new HashSet<String>();
//		sourceOutputFields2.add("seq2");
//		sourceOutputFields2.add("integer2");
//		RATreeNode sourceNode2 = new SourceNode(
//				null, 
//				sourceOutputFields, 
//				"test", "kafka", null);
//
//		RATreeNode joinNode = new JoinNode(
//				new Condition(
//						new FieldValue(
//								"integer", 
//								false), 
//								new FieldValue(
//										"integer2", 
//										false), 
//										Operator.EQUAL), 
//										sourceOutputFields, sourceOutputFields);
//
//		joinNode.addChild(selectNode);
//		joinNode.addChild(sourceNode2);
//
//
//		RATreeNode root = new WindowNode(2, WindowType.SECONDS, 10);
//		root.addChild(joinNode);
//
//		StormTopology topology = topologyFactory.makeTopology(root);
//
//		LocalCluster cluster = new LocalCluster();
//		cluster.submitTopology("test", conf, topology);
//		Utils.sleep(20000);
//		cluster.killTopology("test");
//		cluster.shutdown();
//
//		System.err.println("Done");

	}

}
