package raincoat.ratree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import backtype.storm.generated.Grouping;

import raincoat.api.OperationCost;
import raincoat.api.RATreeNode;
import raincoat.api.RaincoatGrouping;
import raincoat.sql.AlgebraOperator;

public abstract class Node implements RATreeNode {

	private List<RATreeNode> children;
	private AlgebraOperator operationType;
	private List<String> outputTuples;
	private String _identifier;
	private Boolean _needsPostProcessing;
	private RATreeNode _postProcessingNode;
	private Integer _parallelismHint;
	private RATreeNode _parent;

	public Node(AlgebraOperator operationType, List<String> outputTuples) {
		this(UUID.randomUUID().toString(), operationType, outputTuples);
	}

	public Node(String identifier, AlgebraOperator operationType,
			List<String> outputTuples) {
		this.children = new ArrayList<RATreeNode>();
		this.operationType = operationType;

		this.outputTuples = outputTuples;

		_identifier = identifier;
		
		_needsPostProcessing = false;
		_postProcessingNode = null;
		_parallelismHint = 1;
	}

	@Override
	public String getIdentifier() {
		return _identifier;
	}

	public List<String> getInputTuples() {
		List<String> inputFields = 
				new ArrayList<String>();
		for (RATreeNode child : getChildren()) {
			for (String childOutField : 
				child.getOutputTuples()) {
				inputFields.add(childOutField);
			}
		}
		return inputFields;
	}

	public List<String> getOutputTuples() {
		if (getInputTuples().size() == 0) {
			return outputTuples;
		}
		List<String> input = getInputTuples();
		List<String> output = new ArrayList<String>();
		output.addAll(input);
		output.addAll(outputTuples);
		return output;
	}

	public boolean containsField(String field) {
		for (String tupleName : outputTuples) {
			if(tupleName.equalsIgnoreCase(field))
				return true;
		}
		return false;
	}

	public RATreeNode addChild(RATreeNode node) {
		if (this.children.contains(node)) {
			throw new IllegalArgumentException("Node already added to this node");
		}
		this.children.add(node);
		node.setParent(this);

		return this;
	}
	
	
	public boolean removeChild(RATreeNode node) {
		if (this.children.remove(node)) {
			if (node.getParent() == this)
				node.setParent(null);
			return true;
		}
		return false;
	}
	
	public RATreeNode getParent() {
		return _parent;
	}
	
	public void setParent(RATreeNode parent) {
		if (parent == null) {
//			if (_parent.getChildren().contains(this))
//				_parent.removeChild(this);
		}
		_parent = parent;
	}

	public List<RATreeNode> getChildren() {
		return Collections.unmodifiableList(this.children);
	}

	@Override
	public AlgebraOperator getOperator() {
		return this.operationType;
	}

	@Override
	public Double getCost(int numTuples) {
		return OperationCost.getCost(this.operationType, numTuples);	
	}

	@Override
	public String toString() {
		String outputTuples = "";
		for (String field : getOutputTuples()) {
			outputTuples += field + ", ";
		}
		return this.operationType + ": " + outputTuples;
	}

	protected List<RATreeNode> getSources() {
		List<RATreeNode> sources = new ArrayList<RATreeNode>();
		for (RATreeNode child : getChildren()) {
			getSources(child, sources);
		}
		return sources;
	}

	private void getSources(RATreeNode currentNode, List<RATreeNode> sources) {
		if(currentNode instanceof SourceNode)
			sources.add(currentNode);

		for (RATreeNode node : currentNode.getChildren()) {
			getSources(node, sources);
		}
	}

	public boolean needsPostProcessing() {
		return _needsPostProcessing;
	}

	public RATreeNode getPostProcessingNode() {
		return _postProcessingNode;
	}
	
	public void setPostProcessingNode(RATreeNode node) {
		_postProcessingNode = node;
	}

	public Integer getParallelismHint() {
		return _parallelismHint;
	}
	
	public void setParallelismHint(int parallelism) {
		_parallelismHint = parallelism;
	}

	/** The grouping used on incoming streams **/
	public RaincoatGrouping getGrouping() {
		return RaincoatGrouping.SHUFFLE;
	}
	
}