package raincoat.ratree;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import backtype.storm.tuple.Tuple;
import raincoat.api.CostModel;
import raincoat.api.RATreeNode;
import raincoat.sql.Condition;
import raincoat.storm.Tuples;

public class RATreeCostModel implements CostModel {

	private Tuples _tuples;
	private double _networkCostConstant;
	public RATreeCostModel(Tuples tuples, double networkCostConstant) {
		_tuples = tuples;
		_networkCostConstant = networkCostConstant;
	}

	@Override
	public Double costOf(RATreeNode rootNode) {
		
		Double predecessorCost = new Double(0.0);
		Integer numTuples = new Integer(0);
		for (RATreeNode child : rootNode.getChildren()) {
			predecessorCost += costOf(child);
			numTuples += tuplesOutputted(child);
		}
		Double operationCost = rootNode.getCost(numTuples) / 
				rootNode.getParallelismHint();
		
		Double networkCost = _networkCostConstant * 
				rootNode.getChildren().size() * 
				rootNode.getParallelismHint();
		
		return predecessorCost + operationCost + networkCost;
	}
	
	private Map<String, Integer> __cache; 
	private Map<String, Integer> _getCache() {
		if (__cache == null)
			__cache = new HashMap<String, Integer>();
		return __cache;
	}
	
	public void resetCache() {
		__cache = new HashMap<String, Integer>();
	}
	
	private Integer tuplesOutputted(RATreeNode childNode) {
		if (_getCache().containsKey(childNode.getIdentifier())) {
			return _getCache().get(childNode.getIdentifier());
		}
		
		int expectedIncomingTuples = 0;
		
		if (childNode instanceof SourceNode) {
			Integer result = filterTuples(childNode.getIdentifier(), 
					_tuples).size();
			_getCache().put(childNode.getIdentifier(), result);
			return result;
		} else {
			for (RATreeNode childOfChild : childNode.getChildren()) {
				expectedIncomingTuples += tuplesOutputted(childOfChild);
			}

			if (childNode instanceof SelectNode) {
				String sourceIdentifier = childNode.getSourceIdentifier().get(0);
				Double passingProbability = selectProbability(
						filterTuples(sourceIdentifier, _tuples), childNode);

				int tuplesPassingFilters = 0;
				for (RATreeNode childOfChild : childNode.getChildren()) {
					tuplesPassingFilters += tuplesOutputted(childOfChild);
				}
				
				Integer result = (int) (passingProbability * tuplesPassingFilters);
				_getCache().put(childNode.getIdentifier(), result);
				return result;

			}
			else if (childNode instanceof JoinNode) {;
				Integer result = expectedNumOfTuples((JoinNode) childNode, expectedIncomingTuples);
				_getCache().put(childNode.getIdentifier(), result);
				return result;
			}
		}
		
		return 0;
	}

	public Double selectProbability(RATreeNode node) {
		Tuples tuples = filterTuples(node.getSourceIdentifier().get(0), 
				_tuples);
		return selectProbability(tuples, node);
	}
	
	public Double selectProbability(Tuples tuples, RATreeNode node) {
		// TODO: implement selectivity for JOIN
		Double passed = 0.0;
		if (!(node instanceof SelectNode)) {
			return 1.0;
		}
		List<Condition> conditions = ((SelectNode) node).getConditions();
		for (Tuple t : tuples) {
			for (Condition cond : conditions) {
				if (cond.compare(t)) {
					passed++;
				}
			}
		}
		return passed / tuples.size();
	}
	
	public Double joinProbability(String source1, String source2, 
			RATreeNode node) {
		Tuples tuplesSource1 = filterTuples(source1, _tuples);
		Tuples tuplesSource2 = filterTuples(source2, _tuples);
		return joinProbability(tuplesSource1, tuplesSource2, node);
	}
	
	public Double joinProbability(Tuples source1, Tuples source2, 
			RATreeNode node) {
		Condition joinCondition = ((JoinNode) node).getCondition();
		
		// calculate the probability that a tuple 
		// from source1 and 2 will joins
		int joinCount = 0;
		for (Tuple t : source1) {
			for (Tuple s : source2) {
				if (isJoinable(s, t, joinCondition)) {
					joinCount++;
				}
			}
		}
		
		double joinProbability = (double)joinCount / 
				(source1.size() * source2.size());
		
		return joinProbability;	
	}
	
	
	private boolean isJoinable(Tuple s, Tuple t, Condition condition) {
		Tuple u = s;
		Tuple v = t;
		if (t.contains((String)condition.getFieldA().getValue())) {
			u = t;
			v = s;
		}
		return condition.compare(u, v);
	}
	
	
	
	public Integer expectedNumOfTuples(JoinNode node, int inputSize) {
		Tuples source1 = filterTuples(node.getSourceIdentifier().get(0), _tuples);
		Tuples source2 = filterTuples(node.getSourceIdentifier().get(1), _tuples);
		
		double joinProbability = joinProbability(source1, 
				source2, node);
		
		int expectedNumOfTuples = (int)Math.ceil( 
				// assume inputSize = window size. 
				// then the expect amount of tuples are 
				// the number joinProbability times the inputSize ? 
				joinProbability * Math.pow(inputSize,2));
//				joinProbability * inputSize);
		return expectedNumOfTuples;
	}
	
	
	public Tuples filterTuples(String source, Tuples tuples) {
		Tuples filtered = new Tuples(); 
		for (Tuple t : _tuples) {
			if (t.getSourceComponent().equals(source)) {
				filtered.add(t);
			}
		}
		return filtered;
	}

}
