package raincoat.ratree.optimizers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import backtype.storm.tuple.Fields;
import backtype.storm.tuple.Values;

import raincoat.api.RATreeNode;
import raincoat.api.TreeOptimizer;
import raincoat.builder.PipelineTuple;
import raincoat.ratree.JoinNode;
import raincoat.ratree.JoinNodeComparator;
import raincoat.ratree.RATreeCostModel;
import raincoat.ratree.SourceNode;
import raincoat.sql.RATreeBuilder;
import raincoat.sql.SqlParser;
import raincoat.sql.SqlScan;
import raincoat.storm.Tuples;
import raincoat.test.TestDataFactory;
import raincoat.util.DFS;

public class OptimizeLeftDeepJoinOrder implements TreeOptimizer {
	
	private RATreeCostModel _costModel;
	public OptimizeLeftDeepJoinOrder(
			RATreeCostModel costModel) {
		_costModel = costModel;
	}

	@Override
	public RATreeNode optimize(RATreeNode node) {
		OperatorsAsList extractJoins 
			= new OperatorsAsList();
		extractJoins.traverserTree(node);
		List<RATreeNode> joinNodes 
			= new ArrayList<RATreeNode>(
					extractJoins.getJoinNodeList());
		
		List<RATreeNode> orderedJoinNodes 
			= new ArrayList<RATreeNode>(joinNodes);
	
		Collections.sort(orderedJoinNodes, 
				new JoinNodeComparator(_costModel));
		
		return rebuildTree(orderedJoinNodes, 
				extractJoins._sources, 
				extractJoins._allOperators);
		
		
	}

	
	private RATreeNode rebuildTree(List<RATreeNode> list, 
			List<RATreeNode> sources, List<RATreeNode> all) {
		
		// 1
		RATreeNode last = list.remove(0);
		last.setParent(null);
		RATreeNode childToRemove = null;
		List<String> sourceIds = 
				new ArrayList<String>(last.getSourceIdentifier());
		
		for (RATreeNode child : last.getChildren()) {
			if (child instanceof JoinNode) {
				childToRemove = child;
				break;
			}
		}
		last.removeChild(childToRemove);
		
		String source = "";
		for (String sourceId : sourceIds) {
			if (!last.getSourceIdentifier().contains(sourceId)) {
				source = sourceId;
			}
		}
		
		//2
		for (RATreeNode aSource : sources) {
			if (aSource.getIdentifier()
					.equals(source)) {
				last.addChild(aSource);
				break;
			}
		}
		
		// 3 for the rest we must choose the 
		// one with the lowest selectivity that fits
		
		while (list.size() > 0) {
			RATreeNode next = null;
			for (RATreeNode candidate : list) {
				String common = getCommonSource(last, candidate);
				if (common != null) {
					next = candidate;
					break;
				}
			}
			
			list.remove(next);
			pruneChildFromSource(getCommonSource(last, next), 
					next);
			
			next.addChild(last);
			
			last = next;
		}
		
		for (RATreeNode node : all) {
			if (node instanceof JoinNode || 
					node instanceof SourceNode) {
				continue;
			}
			
			if (node.getChildren().size() == 1)
				node.removeChild(node.getChildren().get(0));
			
			node.addChild(last);
			last = node;
		}
		

		return last;
	}
	
	private void pruneChildFromSource(String source, RATreeNode node) {
		RATreeNode remove = null;
		for (RATreeNode child : node.getChildren()) {
			if (child.getSourceIdentifier().contains(source))
				remove = child;
		}
		node.removeChild(remove);
	}
	
	private String getCommonSource(RATreeNode node1, RATreeNode node2) {
		for (String source : node1.getSourceIdentifier()) {
			if (node2.getSourceIdentifier().contains(source)) {
				return source;
			}
		}
		return null;
	}
	
	
	public class OperatorsAsList extends DFS {

		private List<RATreeNode> _joinNodes; 
		private List<RATreeNode> _allOperators;
		private List<RATreeNode> _sources;
		public OperatorsAsList() {
			_joinNodes = new ArrayList<RATreeNode>();
			_allOperators = new ArrayList<RATreeNode>();
			_sources = new ArrayList<RATreeNode>();
		}
		
		public List<RATreeNode> getJoinNodeList() {
			return Collections.unmodifiableList(_joinNodes);
		}
		
		@Override
		protected void execute(RATreeNode root) {
			if (root instanceof JoinNode) {
				_joinNodes.add(root);
			}
			else if (root instanceof SourceNode) {
				_sources.add(root);
			}
			_allOperators.add(root);
		}
	}
	
	
	public static void main(String[] args) {
		
		query("CREATE stream S " +
				"WITH FIELDS s1, s2 " +
				"TYPE randomint " +
				"OPTIONS from=0, to=10, rate=200");
		
		query("CREATE stream T " +
				"WITH FIELDS t1, t2 " +
				"TYPE randomint " +
				"OPTIONS from=0, to=10, rate=200");
		
		query("CREATE stream U " +
				"WITH FIELDS u1, u2 " +
				"TYPE randomint " +
				"OPTIONS from=0, to=10, rate=200");
		
		query("CREATE stream V " +
				"WITH FIELDS v1, v2 " +
				"TYPE randomint " +
				"OPTIONS from=0, to=10, rate=200");
		
		RATreeNode query = query(
				"select s1,t1 from S,T,U,V " +
				"where s1 = t1 " +
				"and s1 = u1 " +
				"and s1 = v1 " +
				"and s1 > 10 " +
				"every 2 seconds size 10");
		
		RATreeCostModel costModel = TestDataFactory.makeCostModel(query);
		
		OptimizeLeftDeepJoinOrder joinOrderOptimization 
		= new OptimizeLeftDeepJoinOrder(costModel);
		
		RATreeBuilder.printDebug(joinOrderOptimization.optimize(query));
		
	}
	
	
	public static RATreeNode query(String query) {
		SqlParser parser = new SqlParser(new SqlScan(query));
		return parser.parse();
	}

	
	
	

}
