package raincoat.ratree.optimizers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import raincoat.api.RATreeNode;
import raincoat.api.TreeOptimizer;
import raincoat.builder.RaincoatFactory;
import raincoat.ratree.AggregatorNode;
import raincoat.ratree.RATreeCostModel;
import raincoat.ratree.SelectNode;
import raincoat.ratree.SourceNode;
import raincoat.sql.RATreeBuilder;
import raincoat.test.TestDataFactory;
import raincoat.util.DFS;

public class OptimizeSelectOrdering implements TreeOptimizer {
	
	private RATreeCostModel _costModel;
	private boolean moveOnly;
	public OptimizeSelectOrdering(RATreeCostModel costModel, boolean moveOnly) {
		_costModel = costModel; 
		this.moveOnly = moveOnly;
	}

	@Override
	public RATreeNode optimize(RATreeNode node) {
		DetectSelectNodes findSelects = new DetectSelectNodes();
		findSelects.traverserTree(node);
		List<RATreeNode> selects = 
				findSelects.getSelects();
		
		if (!moveOnly) 
			Collections.sort(selects, new SelectComparator(_costModel));
		for (RATreeNode select : 
			new ArrayList<RATreeNode>(selects)) {
		
			String sourceIdentifier = select 
					.getSourceIdentifier().get(0);
			RATreeNode parent = select.getParent();
			RATreeNode child = select.getChildren().get(0);
			parent.addChild(child);
			parent.removeChild(select);	
			select.removeChild(child);
			
			RATreeNode source = getSource(
					sourceIdentifier, findSelects._sources);
			
			
			RATreeNode newParent = source.getParent();
			newParent.removeChild(source);
			select.addChild(source);
			newParent.addChild(select);
			
		}
		
		return node;
	}
	
	private RATreeNode getSource(String sourceIdentifier,
			List<RATreeNode> sources) {
		for (RATreeNode source : sources) {
			if (sourceIdentifier.equals(
					source.getIdentifier())) {
				return source;
			}
		}
		return null;
	}
	
	public class DetectSelectNodes extends DFS {

		private List<RATreeNode> _selects; 
		private List<RATreeNode> _sources;
		public DetectSelectNodes() {
			_selects = new ArrayList<RATreeNode>();
			_sources = new ArrayList<RATreeNode>();
		}
		
		public List<RATreeNode> getSelects() {
			return _selects;
		}
		
		@Override
		protected void execute(RATreeNode root) {
			if (root instanceof SelectNode ||
					root instanceof AggregatorNode) {
				_selects.add(root);
			} 
			else if (root instanceof SourceNode) {
				_sources.add(root);
			}
		}
	}
	
	public static void main(String[] args) {
			
			OptimizeLeftDeepJoinOrder.query(
					"CREATE stream S " +
					"WITH FIELDS s1, s2 " +
					"TYPE randomint " +
					"OPTIONS from=0, to=10, rate=200");
			
			OptimizeLeftDeepJoinOrder.query(
					"CREATE stream T " +
					"WITH FIELDS t1, t2 " +
					"TYPE randomint " +
					"OPTIONS from=0, to=2, rate=200");
			
			OptimizeLeftDeepJoinOrder.query(
					"CREATE stream U " +
					"WITH FIELDS u1, u2 " +
					"TYPE randomint " +
					"OPTIONS from=0, to=10, rate=200");
			
			OptimizeLeftDeepJoinOrder.query(
					"CREATE stream V " +
					"WITH FIELDS v1, v2 " +
					"TYPE randomint " +
					"OPTIONS from=0, to=10, rate=200");
			
			RATreeNode query = OptimizeLeftDeepJoinOrder.query(
					"select s1,t1 from S,T,U,V " +
					"where s1 = t1 " +
					"and s1 = u1 " +
					"and s1 = v1 " +
					"and s1 > 10 " +
					"and s2 < 10 " + 
					"and v1 = 5 " +
					"every 2 seconds size 10");
			
			RATreeCostModel costModel = TestDataFactory.makeCostModel(query);
			
			Double costOfNaive = costModel.costOf(query.getChildren().get(0));
			
			OptimizeLeftDeepJoinOrder joinOrderOptimization 
			= new OptimizeLeftDeepJoinOrder(costModel);
			
			RATreeNode hello = joinOrderOptimization.optimize(query);
			RATreeBuilder.printDebug(hello);
			
			costModel.resetCache();
			Double costAfterJoinOpt = costModel.costOf(hello.getChildren().get(0));
					
			
			OptimizeSelectOrdering selectOrdering = 
					new OptimizeSelectOrdering(costModel, true);
			
			costModel.resetCache();
			RATreeNode selectOrder = selectOrdering.optimize(hello);
			
			Double costAfterSel = costModel.costOf(selectOrder.getChildren().get(0));
			
			
			MergeSelects merge = new MergeSelects();
			RATreeBuilder.printDebug(selectOrder);
			RATreeNode selectsMerged = merge.optimize(selectOrder);
			RATreeBuilder.printDebug(selectsMerged);
			
			
			System.err.println("1. Cost of naive: " + costOfNaive);
			System.err.println("2. Cost after join opt: " + costAfterJoinOpt);
			System.err.println("3. Cost after sel opt: " + costAfterSel);
			System.err.println("Cost diff 1. and 2: " + (costOfNaive - costAfterJoinOpt));
			System.err.println("Cost diff 1. and 3: " + (costOfNaive - costAfterSel));
			
			
			RaincoatFactory factory = new RaincoatFactory();
			factory.makeTopology(selectOrder);
			

	}
	
	public class SelectComparator implements Comparator<RATreeNode> {

		private RATreeCostModel _costModel;
		private Map<RATreeNode, Double> _cache;
		public SelectComparator(RATreeCostModel costModel) {
			_costModel = costModel;
			_cache = new HashMap<RATreeNode, Double>();
		}
		@Override
		public int compare(RATreeNode o1, RATreeNode o2) {
			return getSelectivity(o1).compareTo(getSelectivity(o2));
		}
		
		private Double getSelectivity(RATreeNode node) {
			if (_cache.containsKey(node)) 
				return _cache.get(node);
			Double selectivity = _costModel.selectProbability(node);
			_cache.put(node, selectivity);
			return selectivity;
		}
	}

}
