package raincoat.sql;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import raincoat.ratree.Node;
import raincoat.ratree.SourceNode;
import raincoat.ratree.StaticSourceNode;
import raincoat.ratree.WindowNode;
import raincoat.ratree.WindowNode.WindowType;

public class SqlParser {
    private Tokenstream ts; 

    public SqlParser(Tokenstream ts) {
        this.ts = ts;
    }
    
    public Node parse() {
        switch(ts.tok) {
        case Tokenstream.SELECT:
            ts.next();
            return select();
        case Tokenstream.CREATE:
            ts.next();
            return create();
        default:
            throw ts.parseerror("Expected select");
        }
    }
    
    private Node select() {
        List<String> projection_fields = select_expr();
        List<String> join_fields = from();
        List<Condition> conditions = where();
//      group_by();
        WindowNode window = window();
        
        return RATreeBuilder.buildTree(projection_fields, join_fields, conditions, window);
    }

	private WindowNode window() {
		int interval;
		WindowType type;
		int windowSize;
		
		switch(ts.tok) {
		case Tokenstream.EVERY:
			ts.next();
			interval = getInt(ts.sval);
			break;
		default:
			throw ts.parseerror("Expected EVERY");
		}
		
		ts.next();
		switch(ts.tok) {
		case Tokenstream.SECONDS:
			type = WindowType.SECONDS;
			break;
		case Tokenstream.TUPLES:
			type = WindowType.TUPLES;
			break;
		default:
			throw ts.parseerror("Expected SECONDS or TUPLES");
		}
			
		ts.next();
		switch(ts.tok) {
		case Tokenstream.SIZE: 
			ts.next();
			windowSize = getInt(ts.sval);
		break;
		default:
			throw ts.parseerror("Expected WINDOW");
		}
		return new WindowNode(interval, type, windowSize);
	}
	
	private int getInt(String value) {
		try {
			return Integer.parseInt(value);
		} catch(NumberFormatException e) {
			throw ts.parseerror("Expected integer value");
		}
	}

	private List<Condition> where() {
        List<Condition> conditions = new ArrayList<Condition>();
        switch(ts.tok) {
        case SqlScan.WHERE:
            ts.next();
            conditions.add(condition());
            optional_conditions(conditions);
            return conditions;
        default:
            return conditions;
        }
    }
    
    private Condition condition() {
        FieldValue fieldA;
        FieldValue fieldB;
        Condition.Operator operator;
        switch(ts.tok) {
        case Tokenstream.NAME:
            fieldA = field();
            ts.next();
            break;
        case Tokenstream.AGGREGATOR:
        	fieldA = field();
        	ts.next();
        	break;
        case Tokenstream.DATE:
        	ts.next();
        	fieldA = field();
        	ts.next();
        	break;
        default:
            throw ts.parseerror("Expected name attribute");
        }
        
        switch(ts.tok) {
        case Tokenstream.OPERATOR:
            operator = Condition.Operator.fromString(ts.sval);
            ts.next();
            break;
        default:
            throw ts.parseerror("Expected operator");
        }
        
        switch(ts.tok) {
        case Tokenstream.NAME:
            fieldB = field();
            ts.next();
            break;
        case Tokenstream.DATE:
        	ts.next();
        	fieldB = field();
        	ts.next();
        	break;
        default:
            throw ts.parseerror("Expected name attribute");
        }
        return new Condition(fieldA, fieldB, operator);
    }
    
    private void optional_conditions(List<Condition> conditions) {
        switch(ts.tok) {
        case Tokenstream.OPERATION:
            ts.next();
            conditions.add(condition());
            optional_conditions(conditions);
            return;
        case Tokenstream.NAME:
            throw ts.parseerror("Expected operation");
        default:
            return;
        }
    }

    private FieldValue field() {
    	String value = "";
    	if(ts.sval.charAt(0) == '\"' || ts.sval.charAt(0) == '\'') {
    		ts.sval = ts.sval.substring(1);
    		char lastChar = ts.sval.charAt(ts.sval.length() - 1);
    		while(lastChar!= '\"' && lastChar != '\'') {
    			value += " " + ts.sval;
    			ts.next();
    		}
    		value += ts.sval.substring(0, ts.sval.length() - 1);
    		//is date or or string
    		return FieldValue.fromString(value);
    	} 
    	else {
    		if(ts.tok ==Tokenstream.AGGREGATOR) {
    			 String val1 = ts.sval;
    	            ts.next();
    	            val1 += ":" + ts.sval;
    	            //is aggregator
    	            return FieldValue.fromAggregator(val1);
			}
    		// is int or tupleField
    		return FieldValue.fromConstant(ts.sval);
    	}
    }
    
    private List<String> select_expr() {
        List<String> list = new ArrayList<String>();
        String projections = "";
        projections += attribute();
        projections = optional_attributes(projections);
        
        Collections.addAll(list, projections.split(","));
        return list;
    }
    
    private String attribute() {
        switch(ts.tok) {
        case Tokenstream.NAME:
            String val = ts.sval;
            ts.next();
            return val;
        case Tokenstream.AGGREGATOR:
            String val1 = ts.sval;
            ts.next();
            val1 += ":" + ts.sval;
            ts.next();
            return val1;
        default:
            throw ts.parseerror("Expected name attribute or aggregator");
        }
    }
    
    private String optional_attributes(String attributes) {
        switch(ts.tok) {
        case Tokenstream.NAME:
            attributes = attributes + "," + ts.sval;
            ts.next();
            return optional_attributes(attributes);
        case Tokenstream.AGGREGATOR:
            String val1 = ts.sval;
            ts.next();
            val1 += ": " + ts.sval;
            ts.next();
            attributes = attributes + "," + val1;
            return optional_attributes(attributes);
        default:
            return attributes;
        }
    }
    
    private List<String> from() {
        ArrayList<String> list = new ArrayList<String>();
        switch(ts.tok) {
        case Tokenstream.FROM:
            ts.next();
            break;
        default: 
            throw ts.parseerror("Expected FROM");
        }
        String attributes = attribute();
        attributes = optional_attributes(attributes);
        Collections.addAll(list, attributes.split(","));
        
        return list;
    }
    
    private Node create() {
    	boolean isStatic = false;
    	String name;
    	List<String> fields;
    	String type;
    	HashMap<String, String> options =  new HashMap<String, String>();
    	switch(ts.tok) {
    	case Tokenstream.STREAM:
    		ts.next();
    		name = field_name();
    		ts.next();
    		break;
    	case Tokenstream.STATIC:
    		ts.next();
    		isStatic = true;
    		name = field_name();
    		ts.next();
    		break;
		default:
			throw ts.parseerror("expected STREAM");
    	}
		switch(ts.tok) {
		case Tokenstream.WITH:
			ts.next();
			break;
		default:
			throw ts.parseerror("Expected WITH");
		}
		
		switch (ts.tok) {
		case Tokenstream.FIELDS:
			ts.next();
			fields = select_expr();
			break;
		default:
			throw ts.parseerror("Expected FIELDS");
		}
		
		switch(ts.tok) {
		case Tokenstream.TYPE:
			ts.next();
			type = ts.sval;
			ts.next();
			break;
			default:
				throw ts.parseerror("expected TYPE");
		}
		
		switch(ts.tok) {
		case Tokenstream.OPTIONS:
			ts.next();
			options = options();
			
			break;
			default:
				options = new HashMap<String, String>();
				break;
		}
    	System.out.println(options.keySet());
    	System.out.println(options.values());
    	SourceNode stream;
    	if(isStatic)
    		stream = new StaticSourceNode(fields, name, type, options);
    	else 
    		stream = new SourceNode(fields, name, type, options);
    	RATreeBuilder.addSource(stream);
    	return stream;
	}
    
    private HashMap<String, String> options() {
    	HashMap<String, String> opts = new HashMap<String, String>();
    	String[] vals = ts.sval.split("=");
    	opts.put(vals[0], vals[1]);
    	ts.next();
    	opts = optional_options(opts);
    	
    	return opts;
    }
    
    private HashMap<String, String> optional_options(HashMap<String, String> opts) {
    	switch (ts.tok) {
		case Tokenstream.NAME:
			String[] vals = ts.sval.split("=");
	    	opts.put(vals[0], vals[1]);
	    	ts.next();
	    	optional_options(opts);
			break;

		case Tokenstream.EOF:
			break;
		}
    	return opts;
    }
    
    private String field_name() {
    	switch(ts.tok) {
    	case Tokenstream.NAME:
    		return ts.sval;
		default:
			throw ts.parseerror("Expected NAME attribute");
    	}
    }
    
    public static void main(String[] args) {
    	String query = "CREATE static A WITH FIELDS a, b, c, d TYPE kafka OPTIONS ip=10.10.10.10, port=1000";
    	
    	String part = "CREATE STATIC part WITH FIELDS p_partkey, p_name, p_mfgr, p_brand, p_type, p_size, p_container, p_retailprice, p_comment TYPE static";

    	String partsupp = "CREATE STATIC partsupp WITH FIELDS ps_partkey, ps_suppkey, ps_availqty, ps_supplycost, ps_comment TYPE static";
    	
    	String supplier = "CREATE STATIC supplier WITH FIELDS s_suppkey, s_name, s_address, s_nationkey, s_phone, s_acctbal, s_comment TYPE static";

    	String nation = "CREATE STATIC nation WITH FIELDS n_nationkey, n_name, n_regionkey, n_comment TYPE static";

    	String region = "CREATE STATIC region WITH FIELDS r_regionkey, r_name, r_comment TYPE static";

    	String customer = "CREATE STATIC customer WITH FIELDS c_custkey, c_name, c_address, c_nationkey, c_phone, C_acctbal, c_mktsegment, c_comment TYPE static";

    	String lineitem = "CREATE STREAM lineitem WITH FIELDS l_orderkey, l_partkey, l_suppkey, l_linenumber, l_quantity, l_extendedprice, l_discount, l_tax, l_returnflag, l_linestatus, l_shipdate, l_commitdate, l_receipdate, l_shipinstruct, l_shipmode, l_comment TYPE kafka";

    	String orders = "CREATE STREAM orders WITH FIELDS o_orderkey, o_custkey, o_orderstatus, o_totalprice, o_orderdate, o_orderpriority, o_clerk, o_shippriority, o_comment TYPE kafka";
    	
    	String complex1 = "SELECT l_orderkey, o_orderdate, sum(l_extendedprice), o_shippriority FROM customer, orders, lineitem  WHERE c_mktsegment = 'BUILDING' AND c_custkey = o_custkey AND l_orderkey = o_orderkey AND o_orderdate < date '1995-03-15' AND l_shipdate > date '1995-03-15' EVERY 10 SECONDS SIZE 60";
    	
    	new SqlParser(new SqlScan(orders)).parse();
    	new SqlParser(new SqlScan(lineitem)).parse();
    	new SqlParser(new SqlScan(customer)).parse();
    	new SqlParser(new SqlScan(region)).parse();
    	new SqlParser(new SqlScan(nation)).parse();
    	new SqlParser(new SqlScan(supplier)).parse();
    	new SqlParser(new SqlScan(partsupp)).parse();
    	new SqlParser(new SqlScan(part)).parse();
    	
    	new SqlParser(new SqlScan(complex1)).parse();
    	
//    	String query = "CREATE stream A WITH FIELDS a, b, c, d TYPE kafka OPTIONS ip=10.10.10.10, port=1000";
//    	String query = "CREATE stream B WITH TUPLE d, e, f";
//    	String query = "SELECT a, b, c, SUM(a) from streamA, streamB where a = b EVERY 1 SECONDS SIZE 5";
//    	String query = "select SUM(a), b from streamA, streamB where a = b and a = c and a > 10 EVERY 3 SECONDS SIZE 5";
    	
//    	String query2 = "select * from A WHERE a < date '1995-05-15' every 2 SECONDS size 10";
    	String query2 = "SELECT l_orderkey, o_orderdate, sum(l_extendedprice), o_shippriority FROM customer, order, lineitem  WHERE c_mktsegment = 'BUILDING' AND c_custkey = o_custkey AND l_orderkey = o_orderkey AND o_orderdate < date '1995-03-15' AND l_shipdate > date '1995-03-15' EVERY 10 SECONDS SIZE 60";
//    	String query = "select a,b,c from streamA, StreamB every 2 SECONDS size 10";
//    	String query = "select a, b from streamA, streamB where a = b every 5 SECONDS size 10";
//    	String query = "select a, b, c from streamA, StreamB, streamC where b = d and c = d every 2 SECONDS size 10";
//    	String query = "select a, b from streamA, streamB where a > 10 every 2 SECONDS size 10";
//    	String query = "select a, b from streamA, streamB, StreamC where a > 10 and b = d every 2 SECONDS size 10"; 
    	
        new SqlParser(new SqlScan(query)).parse();
//        Node n = new SqlParser(new SqlScan(query2)).parse();
//    	Node n = new SqlParser(new SqlScan("select SUM(b), a from A, B where SUM(a) > 10 and sum(a) > 20")).parse();
//    	System.out.println(n.getOutputTuples());
    }
}
