package raincoat.storm;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.yaml.snakeyaml.emitter.Emitable;

import raincoat.api.Operation;
import raincoat.api.OperationType;
import raincoat.api.Window;
import raincoat.operations.HashJoin;
import raincoat.sql.Condition;
import raincoat.testsuit.performance.BoltStatistics;
import raincoat.windows.HashJoinWindowCondition;
import raincoat.windows.KeySlidingTimedWindow;
import raincoat.windows.SlidingTimedWindow;
import backtype.storm.task.OutputCollector;
import backtype.storm.task.TopologyContext;
import backtype.storm.topology.OutputFieldsDeclarer;
import backtype.storm.topology.base.BaseRichBolt;
import backtype.storm.tuple.Fields;
import backtype.storm.tuple.Tuple;
import backtype.storm.tuple.Values;

// The API we have today made it hard to implement a 
// suitable hashjoin. This implementation is an 
// illustration of what has to change in the 
// Operator interface

public class HashJoinBolt extends BaseRichBolt {
	
	
	private OutputCollector _collector;
	private Operation<Tuple> _operation;
	
	private Integer _maxAge;
	private Integer _tickSize;
	private String _streamA;
	private String _streamB;
	private Window<Tuple> _window;
	private Condition _condition;
	private Fields _outputFields;
	
	private BoltStatistics _boltStats;
	
	public HashJoinBolt(Condition condition, Fields outFields, String streamA, 
			String streamB, Integer maxAge, Integer tickSize, Fields outputFields) {
		_maxAge = maxAge;
		_tickSize = tickSize;
		_streamA = streamA;
		_streamB = streamB;
		_operation = new HashJoin(condition, outFields);
		_condition = condition;
		_outputFields = outputFields;
	}

	@Override
	public void prepare(Map stormConf, TopologyContext context,
			OutputCollector collector) {
		_collector = collector;
		_window = new KeySlidingTimedWindow(new String[]{_streamA, _streamB }, 
				new HashJoinWindowCondition(
						(String)_condition.getFieldA().getValue(),
						(String)_condition.getFieldB().getValue()), 
						_maxAge, _tickSize);
		
		_operation.init(stormConf, _window);
		
		_boltStats = new BoltStatistics("join_"+context.getThisComponentId(), 
				""+ context.getThisTaskId(), 
				5000);
	}

	@Override
	public void execute(Tuple input) {
		long before = System.currentTimeMillis();
		if(input.getSourceStreamId().equals("__tick")) {
			_boltStats.setWindowSize(_window.getTuples().size());
			_window.advanceWindow();
			Tuples evictedTuples = _window.evictExpiredTuples();
			ackTuples(evictedTuples);
			
			_boltStats.incAckedTuples(evictedTuples.size());
			long execTime = System.currentTimeMillis() - before;
			_boltStats.addExecTime(execTime);
			_boltStats.log();
			
			if (_boltStats.ready()) { 
				_boltStats.flush();
			}
			return;
		}
		
		boolean operationPerformed = _operation.performOperation(input);		
		if (operationPerformed) {
			List<Values> output = _operation.getOutputValues();
			for (Values values : output) {
				_collector.emit(_operation.getAnchoringTuples(), values);
			}
			_boltStats.incEmittedTuples(output.size());
		}
		long execTime = System.currentTimeMillis() - before;
		_boltStats.addExecTime(execTime);
		
		
	}
	
	private void ackTuples(Tuples tuples) {
		for (Tuple t : tuples) {
			_collector.ack(t);
		}
	}

	@Override
	public void declareOutputFields(OutputFieldsDeclarer declarer) {
		declarer.declare(_outputFields);
		
	}
	
}
