package raincoat.testsuit.performance;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import backtype.storm.Config;
import backtype.storm.LocalCluster;
import backtype.storm.generated.StormTopology;
import backtype.storm.spout.SchemeAsMultiScheme;
import backtype.storm.spout.SpoutOutputCollector;
import backtype.storm.task.TopologyContext;
import backtype.storm.topology.TopologyBuilder;
import backtype.storm.utils.Utils;

import storm.kafka.KafkaSpout;
import storm.kafka.SpoutConfig;
import storm.kafka.StringScheme;
import storm.kafka.KafkaConfig.StaticHosts;

public class LoggingKafkaSpout extends KafkaSpout {

	private PrintWriter _writer;
	private LinkedList<String> _logCache;
	private long _writeIntervalMs;
	private long _lastUpdated;
	private long _lastTick;
	
	private static String logFormat = "%s %d";
	
	
	/* The data to log  */
	private int __tuplesAcked;
	
	public LoggingKafkaSpout(SpoutConfig spoutConf, 
			long writeIntervalMs) {
		super(spoutConf);
		
		_writeIntervalMs = writeIntervalMs;
		
	}
	
	@Override
	public void open(Map conf, final TopologyContext context, 
			final SpoutOutputCollector collector) {
		super.open(conf, context, collector);
		
		Map<String, String> env = System.getenv();
		initWriter(env.get("HOME") + "/logs/" + 
				context.getThisComponentId() + "_" + context.getThisTaskId() + ".log");
		_logCache = new LinkedList<String>();
		
		__tuplesAcked = 0;
		_lastTick = System.currentTimeMillis();
		
		
	}
	
	private void initWriter(String logfile) {
		try {
			_writer = new PrintWriter(logfile, "UTF-8");
		} catch (FileNotFoundException e) {
			System.err.println("Could not find file: " + logfile);
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		}
	}
	
	@Override
	public void ack(Object ack) {
		super.ack(ack);
		
		log();
	}
	
	private void log() {
		__tuplesAcked++;
		
		if (System.currentTimeMillis() - _lastTick > 2000) {
			_logCache.add(String.format(logFormat, timestamp(),
					__tuplesAcked));
			_lastTick = System.currentTimeMillis();
		}
		
		if (shouldPersistLog()) {
			String toLog = prepareWriteString();
			_writer.print(toLog);
			_writer.flush();
			_logCache = new LinkedList<String>();
			_lastUpdated = System.currentTimeMillis();
		}
	}
	
	private String timestamp(){ 
		
		DateFormat df = new SimpleDateFormat("HH:mm:ss");
		Date today = Calendar.getInstance().getTime();        
		String timestamp = df.format(today);
		return timestamp;
	}
	
	private String prepareWriteString() {
		StringBuilder builder = new StringBuilder();
		for (String line : _logCache) {
			builder.append(line);
			builder.append('\n');
		}
		return builder.toString();
	}
	
	
	private boolean shouldPersistLog() {
		return intervalSinceUpdated() > _writeIntervalMs;
	}
	
	private long intervalSinceUpdated() {
		return System.currentTimeMillis() - 
				getLastUpdated();
	}
	
	private long getLastUpdated() {
		if (_lastUpdated == 0) {
			_lastUpdated = System.currentTimeMillis();
		}
		return _lastUpdated;
	}
	
	public static void main(String[] args) {
		Config conf = new Config();
		conf.setDebug(true);
		conf.setNumWorkers(2);
		conf.put("topology.tick.tuple.freq.secs", 2);
		
		List<String> hostStrings = new ArrayList<String>();
		hostStrings.add("localhost");
		
		SpoutConfig spoutConf = 
				new SpoutConfig(
						StaticHosts.fromHostString(hostStrings, 1), 
						"logging-test", 
						"/kafka-spout", 
						"good-id");
		spoutConf.forceStartOffsetTime(-2);
		spoutConf.scheme = new SchemeAsMultiScheme(new StringScheme());
		LoggingKafkaSpout spout = new LoggingKafkaSpout(spoutConf, 5000);
		
		TopologyBuilder builder = new TopologyBuilder();
		builder.setSpout("logging-kafka-spout-test", spout);
		
		StormTopology topology = builder.createTopology();

		LocalCluster cluster = new LocalCluster();
		cluster.submitTopology("test", conf, topology);
		Utils.sleep(20000);
		cluster.killTopology("test");
		cluster.shutdown();
	}

}
