package raincoat.testsuit.performance;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import raincoat.RaincoatTesting;
import raincoat.api.RATreeNode;
import raincoat.builder.PipelineTuple;
import raincoat.builder.RaincoatFactory;
import raincoat.ratree.RATreeCostModel;
import raincoat.ratree.SourceNode;
import raincoat.ratree.WindowNode;
import raincoat.ratree.optimizers.OptimizeLeftDeepJoinOrder;
import raincoat.ratree.optimizers.OptimizeSelectOrdering;
import raincoat.ratree.optimizers.MergeSelects;
import raincoat.sql.RATreeBuilder;
import raincoat.storm.Tuples;
import raincoat.util.DFS;
import backtype.storm.Config;
import backtype.storm.generated.StormTopology;
import backtype.storm.tuple.Fields;
import backtype.storm.tuple.Tuple;

public class RunQuery {
	
	public static void main(String[] args) {
		
		parseInput(args);
		for (String query : queries) {
			RATreeNode queryTree = RaincoatTesting.parseQuery(query);
			if (queryTree instanceof SourceNode) {
				continue;
			}
			
			if (parallelize) {
				ApplyParallelismHint parallelismHint = new ApplyParallelismHint(parallelism.toArray());
				parallelismHint.traverserTree(queryTree);
			}
			
			
			RATreeCostModel costModel = 
					new RATreeCostModel(getSampleTuplesSet(), 
					0.001);

			// Optimizers
			
			if (optimizers.size() > 0) {
				queryTree = optimize(queryTree, optimizers, costModel);
			} else {
				OptimizeSelectOrdering moveSelects = new OptimizeSelectOrdering(costModel, true);
				queryTree = moveSelects.optimize(queryTree);
			}
			
			RATreeBuilder.printDebug(queryTree);
			
			RaincoatFactory factory = new RaincoatFactory();
			StormTopology topology = factory.makeTopology(queryTree);
			String identifier = UUID.randomUUID().toString().split("-")[0];
			WindowNode windowNode = (WindowNode) queryTree;
			Config config = new Config();
			config.put("topology.tick.tuple.freq.secs", windowNode.getTickInterval());
			config.put(Config.TOPOLOGY_MESSAGE_TIMEOUT_SECS, 30);
			config.setNumWorkers(80);
			RaincoatTesting.deployTopology(topology, identifier, config, false);
			PerformanceTestSuit.sleep(5000); // wait for it to start
			
			PerformanceTestSuit.sleep(runTime);
			RaincoatTesting.killTopology(identifier);
			
			break;
		}
	}
	
	private static List<Object> parallelism;
	private static boolean parallelize;
	private static String[] queries;
	private static List<String> optimizers;
	private static long runTime;
	private static void init() {
		queries = new String[] {};
		optimizers = new ArrayList<String>();
		parallelism = new ArrayList<Object>();
		runTime = 0;
		parallelize = false;
	}
	
	
	public static void parseInput(String[] args) {
		init();
		for (int i = 0; i < args.length; i++) {
			String arg = args[i];
			System.err.println(arg);
			if (arg.split("=").length != 2) {
				queries = Arrays.copyOfRange(args, i, args.length);
				return;
			}
			else {
				String[] split = arg.split("=");
				String option = split[0];
				String parameter = split[1];
				
				if (option.equalsIgnoreCase("p")) {
					String[] paras = parameter.split(",");
					for (int j = 0; j < paras.length; j = j + 2) {
						parallelism.add(paras[j].toLowerCase());
						parallelism.add(Integer.parseInt(paras[j+1]));
					}
					parallelize = true;
				}
					
				else if (option.equalsIgnoreCase("o")) {
					// o=select,merge,join
					String[] opts = parameter.split(",");
					for (String optimizer : opts) {
						optimizers.add(optimizer);
					}
				}
				
				else if (option.equalsIgnoreCase("run")) {
					runTime = Long.parseLong(parameter);
				}
					
			}
		}
	}
	
	private static RATreeNode optimize(RATreeNode node, 
			List<String> optimizers, RATreeCostModel costModel) {
		for (String optimizer : optimizers) {
			if (optimizer.equalsIgnoreCase("join")) {
				OptimizeLeftDeepJoinOrder join = new OptimizeLeftDeepJoinOrder(costModel);
				node = join.optimize(node);
				OptimizeSelectOrdering moveSelects = new OptimizeSelectOrdering(costModel, true);
				node = moveSelects.optimize(node);
			} 
			else if (optimizer.equalsIgnoreCase("select")) {
				OptimizeSelectOrdering select = new OptimizeSelectOrdering(costModel, false);
				node = select.optimize(node);
			} else if (optimizer.equalsIgnoreCase("merge")) {
				OptimizeSelectOrdering moveSelects = new OptimizeSelectOrdering(costModel, true);
				node = moveSelects.optimize(node);
				MergeSelects selectMerger = new MergeSelects();
				node = selectMerger.optimize(node);
			}
			
			else {
				// hmm
			}
		}
		return node;
	}
	
	private static Tuples getSampleTuplesSet() {
		Tuples tuples = new Tuples();
		List<SourceNode> sources = RATreeBuilder.getSources();
		RaincoatFactory factory = new RaincoatFactory();
		for (SourceNode source : sources) {
			Map<String,String> options = source.getOptions();
			String type = options.get("type");
			long seed = Long.parseLong(options.get("seed"));
			DataGenerator gen = 
					factory.getGeneratorForType(type, seed);
			for (int i = 0; i < 400; i++) {
				Tuple tuple = new PipelineTuple(source.getIdentifier(), 
						new Fields(source.getOutputTuples()), 
						gen.next());
				
				tuples.add(tuple);
			}
		}
		return tuples;
	}
	
	public static class ApplyParallelismHint extends DFS {
		
		private List<Object> pair;
		public ApplyParallelismHint(Object...objects) {
			pair = new ArrayList<Object>();
			for (Object obj : objects) {
				pair.add(obj);
			}
		}

		@Override
		protected void execute(RATreeNode root) {
			for (int i = 0; i < pair.size(); i = i + 2) {
				if (root.getIdentifier().matches((String)pair.get(i)+ "(.*)")){
					int parallelismHint = (Integer)pair.get(i+1);
					root.setParallelismHint(parallelismHint);
					// Only 1 match pr tree. Optimize by removing the matched pair
					return;
				}
			}
			
		}
		
	}

}
